#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from __future__ import with_statement
from __future__ import unicode_literals

import os
import stat
import calendar
import io
import json

from dropboxclient import (DropboxClient,) 
from fileaccess import (FileObj, Access)
from pkgaddon.dropbox_config import (APP_KEY, APP_SECRET,)
from dropboxclient import (DropboxClient, InvalidResponseDropboxError)


OPEN_MODE_READ = "r"
OPEN_MODE_WRITE = "w"

SESSION_PATH_PERM = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP |
                        stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)


class DropboxFileObj(FileObj):

    def _multi_upload(self):

        if self._offset:
            if self._offset + self._io_bytes.tell() < self._length:
                self._io_bytes.seek(0)
                self._client._call_files_upload_session_append_v2(
                                            self._session_id,
                                            self._offset,
                                            self._io_bytes.read())
            else:
                self._io_bytes.seek(0)
                self._client._call_files_upload_session_finish(
                                            self._session_id,
                                            self._offset,
                                            self._io_bytes.read(),
                                            self._path, mode="overwrite")
                self._session_id = None
            return

        self._io_bytes.seek(0)
        response = self._client._call_files_upload_session_start(
                                            self._io_bytes.read())
        self._session_id = response.get('session_id')
        if self._session_id is None:
            raise InvalidResponseDropboxError('session_id is not exists.')

    def __init__(self, client, path, mode, length):

        super(DropboxFileObj, self).__init__()
        self._client = client
        self._path = path
        self._file = None
        self._mode = mode
        self._length = length
        self._offset = 0
        self._session_id = None
        self._do_multi_upload = False
        if length > self._client._max_upload_size:
            self._do_multi_upload = True

        self._io_bytes = io.BytesIO()

        if (not self._mode == OPEN_MODE_READ and
            not self._mode == OPEN_MODE_WRITE):

            raise ValueError(ERR_MSG_INVALID_OPEN_MODE)

    def close(self):

        if self._mode == OPEN_MODE_READ:
            return

        if self._io_bytes is None:
            return

        self._io_bytes.seek(0)
        if self._do_multi_upload:
            if self._session_id:
                self._client._call_files_upload_session_finish(
                                            self._session_id,
                                            self._io_bytes.read(),
                                            self._path, mode="overwrite")
        else:
            self._client._call_files_upload(self._io_bytes.read(),
                                            self._path, mode="overwrite")
        self._io_bytes = None

    def read(self, size=-1):
        if self._file is None:
            self._file = self._client.get_file(self._path)

        return self._file.read(size)

    def write(self, data):
        """Write a "data" data to current position in file."""

        if self._io_bytes is None:
            self._io_bytes = io.BytesIO()

        self._io_bytes.write(data)

        if self._do_multi_upload:
            if (self._offset + self._io_bytes.tell() < self._length and
                self._io_bytes.tell() < self._client._max_upload_size):

                return len(data)

            self._multi_upload()
            self._offset += self._io_bytes.tell()
        else:
            if self._io_bytes.tell() < self._length:
                return len(data)
            self._io_bytes.seek(0)
            self._client._call_files_upload(self._io_bytes.read(),
                                                self._path, mode="overwrite")
        self._io_bytes = None
        return len(data)


class DropboxAccess():

    def _get_proxy(self, url):

         return self._proxy

    def _to_utf8(self, target, strip_separator=True):

        target = target.rstrip(self.sep)
        if isinstance(target, unicode):
            return target.encode("utf8")
        return target

    def _to_unicode(self, target):

        if isinstance(target, str):
            return target.decode("utf8")
        return target

    def _save_tokens(self, tokens):
        path = self._session_path
        with open(path, "w") as f:
            json.dump(tokens, f)
        os.chmod(path, SESSION_PATH_PERM)

    def __init__(self, app_key, app_secret, session_path=None, proxy=None):
        self.sep = "/"

        self._client = DropboxClient(
            app_key,
            app_secret,
            callbacks={"onobtaintokens": lambda tokens: self._save_tokens(tokens)})
        self._client.get_proxy = self._get_proxy
        self._session_path = session_path
        self._proxy = proxy

    @property
    def max_num_transfer_threads(self):
        return 5

    @property
    def auth_url(self):
        """Get the auth url."""

        return self._client.get_authorize_url()

    def authenticate(self, params=None):
        """Takes in params, creates a session.

        Args:
            params(dict):
                The GET parameter sent to redirect_url
                If session is already valid, is None.
                e.g.
                {'code': '....'}
        """

        if params is None:
            params = {}

        tokens = {}
        if params.get('code') is not None:
            tokens = self._client.obtain_access_token(params['code'])
        else:
            with open(self._session_path, "r") as session_file:
                data = session_file.read()
                try:
                    tokens = json.loads(data)
                except Exception as e:
                    tokens["access_token"] = data
        self._client.access_token = tokens.get('access_token')
        self._client.refresh_token = tokens.get('refresh_token')

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        self._client.create_folder(self._to_utf8(dir_path))

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""

        self._client.remove_file(self._to_utf8(dir_path))

    def stat_directory(self, dir_path):
        """stat "dir_path" directory."""

        resource = self._client.get(self._to_utf8(dir_path))
        try:
            if resource.is_folder:
                return {
                    stat.ST_MODE: stat.S_IFDIR,
                    stat.ST_MTIME: 0,
                    stat.ST_SIZE: None,
                }
        except:
            pass

        return {
            stat.ST_MODE: 0,
            stat.ST_MTIME: (calendar.timegm(resource.modified.timetuple()) -
                                        (9 * 3600)),
            stat.ST_SIZE: resource.size,
        }

    def rename_directory(self, from_path, new_name):

        self._client.move(self._to_utf8(from_path), self._to_utf8(new_name))

    def move_directory(self, from_path, to_path):

        self._client.move(self._to_utf8(from_path), self._to_utf8(to_path))

    def get_contents(self, dir_path, with_stat=False):

        if dir_path == self.sep:
            dir_path = ""
        contents = {}
        for r in self._client.get_files(self._to_utf8(dir_path)):
            p = getattr(r, 'path')
            f = getattr(r, 'is_folder', None)
            m = getattr(r, 'modified', None)
            s = getattr(r, 'size', None)
            name = self._to_unicode(p.split(self.sep)[-1])
            if with_stat:
                if f:
                    contents[name] = {
                        stat.ST_MODE: stat.S_IFDIR,
                        stat.ST_MTIME: 0,
                        stat.ST_SIZE: None,
                    }
                else:
                    contents[name] = {
                        stat.ST_MODE: 0,
                        stat.ST_MTIME: (calendar.timegm(m.timetuple()) -
                                                    (9 * 3600)),
                        stat.ST_SIZE: s,
                    }
            else:
                contents[name] = {}
        return contents

    def open(self, file_path, mode="r", length=-1, chunk_size=None):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        return DropboxFileObj(self._client, file_path, mode, length)

    def delete_file(self, file_path):

        self._client.remove_file(self._to_utf8(file_path))

    def stat_file(self, file_path):

        resource = self._client.get(self._to_utf8(file_path))
        try:
            if resource.is_folder:
                return {
                    stat.ST_MODE: stat.S_IFDIR,
                    stat.ST_MTIME: 0,
                    stat.ST_SIZE: None,
                }
        except:
            pass

        return {
            stat.ST_MODE: 0,
            stat.ST_MTIME: (calendar.timegm(resource.modified.timetuple()) -
                                        (9 * 3600)),
            stat.ST_SIZE: resource.size,
        }

    def rename_directory(self, from_path, new_name):

        self._client.move(self._to_utf8(from_path), self._to_utf8(new_name))

    def move_directory(self, from_path, to_path):

        self._client.move(self._to_utf8(from_path), self._to_utf8(to_path))
